package swtpl.marsrover.gui;

import javax.swing.SwingUtilities;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import swtpl.marsrover.map.MapObject;
import swtpl.marsrover.map.MapStateEnum;
import swtpl.marsrover.net.Connection;
import swtpl.marsrover.net.InitHandler;
import swtpl.marsrover.net.MartianEvent;
import swtpl.marsrover.net.MartianHandler;
import swtpl.marsrover.net.ObstacleEvent;
import swtpl.marsrover.net.ObstacleHandler;
import swtpl.marsrover.net.RunContext;
import swtpl.marsrover.net.VehicleEvent;
import swtpl.marsrover.net.VehicleHandler;
import swtpl.marsrover.odometry.Odometry;
import swtpl.marsrover.planner.GoalHandler;
import swtpl.marsrover.planner.JohoPlanner;
import swtpl.marsrover.planner.Planner;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;

public class ClientGui extends JPanel implements Runnable, VehicleHandler, ObstacleHandler, MartianHandler, GoalHandler {
	
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private int mapHeight;
	private int mapWidth;
	private int baseX;
	private int baseY;
	private int roverX;
    private int roverY;
    private int roverDir;
    private int roverVisionMin;
    private int roverVisionMax;
    private boolean isRoverInitialized = false;
    
    private ArrayList<MapObject> craterList = new ArrayList<MapObject>();
    private ArrayList<MapObject> boulderList = new ArrayList<MapObject>();
    private ArrayList<Point> martianList = new ArrayList<Point>();
    private ArrayList<MapObject> goals = new ArrayList<MapObject>();

    private static final int SCALE = 300;

    //private static final Color ROVER_VISION_OUTER_COLOR = new Color(255, 255, 0);
    //private static final Color ROVER_VISION_COLOR = new Color(255, 255, 255);
    private static final Color ROVER_VISION_COLOR = new Color(255, 255, 0);
    private static final Color BASE_COLOR = new Color(0, 250, 0);
    private static final Color ROVER_COLOR = new Color(83, 25, 147);
    private static final Color ROVER_VISION_DIR_COLOR = new Color(83, 25, 0);
    private static final Color CRATER_COLOR = new Color(147, 82, 0);
    private static final Color BOULDER_COLOR = new Color(120, 120, 120);
    private static final Color MARTIAN_COLOR = new Color(255, 37, 0);

	//test it
    public static void main(String[] args) {
    	if( args.length != 2 ) {
    		System.err.println("USAGE: ClientGui <hostname> <port>");
    		System.exit(-1);
    	}

    	try {
			final Connection con = new Connection(args[0], Integer.parseInt(args[1]));
			final Planner planner = new JohoPlanner(con);
			Odometry o = new Odometry();
			con.registerVehicleHandler(o);
			con.registerInitHandler(new InitHandler() {

				private boolean initialized = false;

				@Override
				public void initialized(RunContext rc) {
					if( initialized )
						return;
					initialized = true;
					ClientGui cg = new ClientGui(con,rc.getCardSizeY(), rc.getCardSizeX(), rc.getMaxSensorMM(), rc.getMinSensorMM());
					con.registerMartianHandler(cg);
					con.registerObstacleHandler(cg);
					con.registerVehicleHandler(cg);
					planner.registerGoalHandler(cg);
				}

			});
			con.startReading(con.getInputStream());
		} catch (NumberFormatException e){
			e.printStackTrace(System.err);
		} catch (UnknownHostException e) {
			e.printStackTrace(System.err);
		} catch (IOException e) {
			e.printStackTrace(System.err);
		}
    }

    Connection connection;

    //all parameters are rover coordinates, not java coordinates
    public ClientGui(Connection con,int mapHeight, int mapWidth, int roverVisionMax, int roverVisionMin){
    	this.mapHeight = mapHeight;
    	this.mapWidth = mapWidth;
    	this.baseX = 0;
    	this.baseY = 0;
        this.roverVisionMin = roverVisionMin;
        this.roverVisionMax = roverVisionMax;
        this.connection = con;

        //init gui
    	SwingUtilities.invokeLater(this);
    }

    //convert rover coordinates to java coordinates
    private int toJavaX(int roverX){
    	return (this.mapWidth/2 + roverX) / SCALE;
    }
    private int toJavaY(int roverY){
    	return (this.mapHeight/2 - roverY) / SCALE;
    }

	public void run() {
        JFrame f = new JFrame("Map " + this.mapHeight + "x" + this.mapWidth);
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        JScrollPane jScrollPane = new JScrollPane(this);
        jScrollPane.setPreferredSize(new Dimension(Math.min(1000, this.mapWidth / SCALE + 20), Math.min(700, this.mapHeight / SCALE +20)));
        f.add(jScrollPane);
        f.pack();
        f.setVisible(true);
        f.addKeyListener(new KeyAdapter() {
        	public void keyPressed(KeyEvent e) {
    			switch( e.getKeyCode() ) {
    			case KeyEvent.VK_LEFT:
    			case KeyEvent.VK_KP_LEFT:
    				connection.sendCommand(Connection.AccelerationCommand.NONE,Connection.TurningCommand.LEFT);
    				break;
    			case KeyEvent.VK_RIGHT:
    			case KeyEvent.VK_KP_RIGHT:
    				connection.sendCommand(Connection.AccelerationCommand.NONE,Connection.TurningCommand.RIGHT);
    				break;
    			case KeyEvent.VK_UP:
    			case KeyEvent.VK_KP_UP:
    				connection.sendCommand(Connection.AccelerationCommand.ACCELERATE,Connection.TurningCommand.NONE);
    				break;
    			case KeyEvent.VK_DOWN:
    			case KeyEvent.VK_KP_DOWN:
    				connection.sendCommand(Connection.AccelerationCommand.BREAK,Connection.TurningCommand.NONE);
    				break;
    			default:
    				break;
    			}
        	}
        });

    }

    public Dimension getPreferredSize() {
        return new Dimension(this.mapHeight/SCALE, this.mapWidth/SCALE);
    }

    public void fillCircle(Graphics g, int x, int y, int radius) {
    	g.fillOval(toJavaX(x-radius), toJavaY(y+radius), 2*radius/SCALE, 2*radius/SCALE);
    }

    public void drawCircle(Graphics g, int x, int y, int radius) {
    	g.drawOval(toJavaX(x-radius), toJavaY(y+radius), 2*radius/SCALE, 2*radius/SCALE);
    }

    //does the actual painting
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        if(!this.isRoverInitialized){
        	return;
        }

        //Draw rover view, always draw this first
        Graphics2D g2 = (Graphics2D)g;
        g2.rotate(-2*Math.PI*roverDir/3600, toJavaX(roverX), toJavaY(roverY));
        g.setColor(ClientGui.ROVER_VISION_COLOR);
        double ellipseHalfHeight = Math.sqrt(roverVisionMin*roverVisionMax);
        
        g2.fillOval(toJavaX(roverX - roverVisionMin), toJavaY(roverY + (int)ellipseHalfHeight), (roverVisionMin + roverVisionMax)/SCALE, (int)(2*ellipseHalfHeight/SCALE));
        g.setColor(ClientGui.ROVER_VISION_DIR_COLOR);
        g2.drawLine(toJavaX(roverX), toJavaY(roverY), toJavaX(roverX + roverVisionMax), toJavaY(roverY));
        g2.rotate(2*Math.PI*roverDir/3600, toJavaX(roverX), toJavaY(roverY));
        
        /*
        double cos = Math.cos(2*Math.PI*roverDir/3600);
        double sin = Math.sin(2*Math.PI*roverDir/3600);
        
        g.setColor(ClientGui.ROVER_VISION_OUTER_COLOR);
        fillCircle(g, roverX, roverY, roverVisionMax);
        g.setColor(ClientGui.ROVER_VISION_COLOR);
        fillCircle(g, roverX, roverY, roverVisionMin);
        g.setColor(ClientGui.ROVER_VISION_DIR_COLOR);
        g.drawLine(toJavaX(roverX - (int)(cos*roverVisionMin)), toJavaY(roverY - (int)(sin*roverVisionMin)),
        		toJavaX(roverX + (int)(cos*roverVisionMax)), toJavaY(roverY + (int)(sin*roverVisionMax)));
        */
        
        //Draw the home base, 5m, centered
        g.setColor(ClientGui.BASE_COLOR);
        fillCircle(g, this.baseX, this.baseY, 5000);

        //Draw the craters
        g.setColor(ClientGui.CRATER_COLOR);
        for(MapObject crater : this.craterList){
            fillCircle(g, crater.getX(), crater.getY(), crater.getRadius());
        }

        //Draw the boulders
        g.setColor(ClientGui.BOULDER_COLOR);
        for(MapObject boulder : this.boulderList){
            fillCircle(g, boulder.getX(), boulder.getY(), boulder.getRadius());
        }

        //Draw the goals
        g.setColor(ClientGui.MARTIAN_COLOR);
        for(MapObject goal : this.goals){
            drawCircle(g, goal.getX(), goal.getY(), goal.getRadius());
        }

        //Draw the rover, 0.5m
        g.setColor(ClientGui.ROVER_COLOR);
        fillCircle(g, roverX, roverY, 500);
        
        //focus on rover
        this.scrollRectToVisible(new Rectangle(toJavaX(this.roverX-this.roverVisionMax), toJavaY(this.roverY+this.roverVisionMax),
        			this.roverVisionMax*2/SCALE, this.roverVisionMax*2/SCALE));

        //Draw the martians, 0.4m
        g.setColor(ClientGui.MARTIAN_COLOR);
        for(Point martian : this.martianList){
        	fillCircle(g, martian.x, martian.y, 400);
        }
    }

	@Override
	public void changed(VehicleEvent ve) {
		this.isRoverInitialized = true;
		martianList.clear();
		roverX = ve.getPosX();
		roverY = ve.getPosY();
		roverDir = ve.getDir();
		repaint();
	}

	@Override
	public void recognized(ObstacleEvent oe) {
		MapObject mo = new MapObject(oe.getX(),oe.getY(),oe.getRadius());
		if( oe.getSort() == MapStateEnum.BOULDER )
			boulderList.add(mo);
		else
			craterList.add(mo);
	}

	@Override
	public void recognized(MartianEvent me) {
		Point pt = new Point(me.getX(),me.getY());
		martianList.add(pt);
	}
	
	
	public void goals(ArrayList<MapObject> goals) {
		this.goals = goals;
	}
}

