package swtpl.marsrover.planner;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Stack;

import swtpl.marsrover.map.MapObject;
import swtpl.marsrover.net.Connection;
import swtpl.marsrover.net.MartianHandler;
import swtpl.marsrover.net.ObstacleEvent;

public class JohoPlanner extends Planner {
	private final static int SAFEDIST = 1000;
	private final static int MINDIST = 5000;
	private final static int PLAN_AHEAD = 50000;
	private final static double STEP = 10;
	private final static double STEP_FACTOR = 1.5;

	class Destination {
		int x;
		int y;
		int radius;
		boolean goRight;
		
		public Destination(int x, int y, int radius, boolean goRight) {
			this.x = x;
			this.y = y;
			this.radius = radius;
			this.goRight = goRight;
		}
		
		public String toString() {
			return "Dest["+x+","+y+":"+radius+(goRight ? "R" : "L")+"]";
		}
		
	}
	
	private Stack<Destination> goals;

	public JohoPlanner(Connection con) {
		super(con);
	}

	private void publishGoals() {
		ArrayList<MapObject> ptgoals = new ArrayList<MapObject>();
		for (Destination dest: goals) {
			ptgoals.add(new MapObject(dest.x, dest.y, dest.radius));
		}
		for (GoalHandler gh : goalHandlers)
			gh.goals(ptgoals);
	}
	
	
	private static int normalizeAngle(int angle) {
		if (angle > 1800)
			return angle - 3600;
		if (angle < -1800)
			return angle + 3600;
		return angle;
	}
	
	public boolean checkTurnFeasible() {
		return true;
	}
	
	private double euklidDistSq(int dx, int dy) {
		return (double)dx*dx + (double)dy*dy;
	}

	public void doCleanup() {
		super.doCleanup();
		while (goals.size() > 1)
			goals.pop();
		if (goalHandlers != null)
			publishGoals();
      //  System.err.println("cleaning up");
	}

	public boolean checkObstacle(Destination nextDest, int wantAccel) {
		double a = wantAccel < 0 ? brake : wantAccel > 0 ? accel : 0.0;
		int angAdd;
		double s = speed;
		double d = dir;
		double x = this.x;
		double y = this.y;
		double step = STEP;
		for (int t = 0; t < PLAN_AHEAD; t+= (int)step) {
			step *= STEP_FACTOR;
			double dist = euklidDistSq(nextDest.y - (int)y, nextDest.x - (int)x);
			double realDist = dist - ((double)nextDest.radius*nextDest.radius);

			if (realDist < MINDIST*MINDIST) {
				return false;
			}
			
			
			double destDir = Math.atan2(nextDest.y - (int)y, nextDest.x - (int)x);
			double realDestDirDelta = Math.atan2((double) nextDest.radius, Math.sqrt((double)realDist));
			boolean hardTurn = false;
			if (nextDest.goRight)
				destDir -= realDestDirDelta;
			else
				destDir += realDestDirDelta;
			int delta = normalizeAngle((int) (destDir* 3600 / (2*Math.PI) - d));
			if (Math.abs(delta) < rc.getTenthDegreePerSecsTurn()/5)
				angAdd = 0;
			else if (Math.abs(delta) < 2*rc.getTenthDegreePerSecsHardTurn()/5)
				angAdd = rc.getTenthDegreePerSecsTurn();
			else {
				angAdd = rc.getTenthDegreePerSecsHardTurn();
				hardTurn = true;
			}
			if (delta < 0)
				angAdd = -angAdd;
			
			s += (int) (step * a - drag * step * s*s);
			if (s < 0)
				s = 0;
			x += Math.cos(d*2*Math.PI/3600) * s * step / 1000;
			y += Math.sin(d*2*Math.PI/3600) * s * step / 1000;
			d += step * angAdd / 1000.0;
			if (d > 1800)
				d -= 3600;
			if (d < -1800)
				d += 3600;
			
			//Graphics g = gui.getGraphics();
			//g.setColor(Color.MAGENTA);
			//g.drawLine(gui.toJavaX((int)x),gui.toJavaY((int)y),
			//		gui.toJavaX((int)x),gui.toJavaY((int)y));
			
			ObstacleEvent obstacle = quadtree.getField((int)x, (int)y);
			if (obstacle != null) {
				/* Found an obstacle on the way */
				boolean evadeRight;
				if (hardTurn)
					evadeRight = angAdd > 0;
				else {
					double obstDir = 
						Math.atan2(obstacle.getY() - y, obstacle.getX() - x);
					evadeRight = normalizeAngle((int)(obstDir*2*Math.PI/3600 - d)) < 0;
				}
                /*
				if (goals.size() > 1) {
					goals.pop();
				}
                */
				addGoal(new Destination(obstacle.getX(), obstacle.getY(),
						obstacle.getRadius()+SAFEDIST, evadeRight));
				if (goalHandlers != null)
					publishGoals();
				return true;
			}
		}
		return false;
	}

    private boolean goalRelevant(Destination dest) {
            long xl = x, yl = y;
            long baseDist = xl*xl+yl*yl;
            xl = dest.x; yl = dest.y;
            long destDist = xl*xl+yl*yl;
            return (baseDist>destDist);
    }

    private void addGoal(Destination dest) {
        if(goalRelevant(dest)) {
            goals.push(dest);
        //    System.err.println("pushing goal: "+dest);
        } else {
         //   System.err.println("not pushing goal: "+dest);
        }
    }

    private void popGoal() {
        if(goals.size()>1) {
            Destination dest = goals.pop();
            long xl = x, yl = y;
            long baseDist = xl*xl+yl*yl;
            xl = dest.x; yl = dest.y;
            long destDist = xl*xl+yl*yl;
         //   System.err.println("popping goal: "+dest+"(base: "+baseDist+", dest " + destDist+")");
        }
    }

    private void popIrrelevantGoals() {
        while(goals.size()>1 && !goalRelevant(goals.peek()))
            popGoal();
    }
	
	public void run() {
		synchronized(this) {
			try {
				while (!positionKnown)
					wait();
			} catch (InterruptedException ex) {
			//	// system.err.println("Can't wait for position:"+ex);
				return;
			}
		}

		goals = new Stack<Destination>();
		goals.push(new Destination(0,0,0, false));
		while (true) {
			try {
                for(int i=0; i<5; i++) {
                    Thread.sleep(20);
                    popIrrelevantGoals();
                }
			} catch (InterruptedException ex) {
				// // system.err.println("Can't wait for position:"+ex);
				return;
			}
			
            Destination nextDest = goals.peek();
			int wantAccel, wantTurn;
			double dist = euklidDistSq(nextDest.y - y, nextDest.x - x);
			double realDist = dist - ((double)nextDest.radius*nextDest.radius);

            /*
			if (realDist < MINDIST*MINDIST) {
				if (goals.size() > 1)
					goals.pop();
				continue;
			}
            */
            if (goalHandlers != null)
                publishGoals();
			
			double destDir = Math.atan2(nextDest.y - y, nextDest.x - x);
			double realDestDirDelta = Math.atan2((double) nextDest.radius, Math.sqrt((double)realDist));
			if (nextDest.goRight)
				destDir -= realDestDirDelta;
			else
				destDir += realDestDirDelta;
			int delta = normalizeAngle((int) (destDir * 3600 / (2*Math.PI)) - dir);
			if (Math.abs(delta) < rc.getTenthDegreePerSecsTurn()/4)
				wantTurn = 0;
			else if (Math.abs(delta) < 2*rc.getTenthDegreePerSecsHardTurn()/5)
				wantTurn = 1;
			else
				wantTurn = 2;
			if (delta < 0)
				wantTurn = -wantTurn;
			
			wantAccel = 1;
			if (wantTurn == 2) {
				if (!checkTurnFeasible()) {
					wantAccel = 0;
					if (!checkTurnFeasible())
						wantAccel = -1;
				}
			}

			if (checkObstacle(nextDest, wantAccel))
				continue;
			driver.setDesiredStearing(wantAccel, wantTurn);
		}
	}
}
